from ..util import Console
from ..interface.IConnection import Base

import psycopg2
import psycopg2.extras

class Connection(Base,object):

    _instances = {}

    def __new__(cls, connectionInfo):
        if not hasattr(cls, 'instance'):
            print('Creating the object...')
            cls.instance = super(Connection, cls).__new__(cls)
        return cls.instance


    def __init__(self, connectionInfo):
        self.connection = self.connect(connectionInfo)
        self.query_parts_og = {}
        self.query_parts = {}


    def connect(self, connectionInfo):
        Console.info('Running connect in connector....')
        conn = psycopg2.connect(database=connectionInfo['database'],
                                host=connectionInfo['host'],
                                user=connectionInfo['user'],
                                password=connectionInfo['password'],
                                port=connectionInfo['port'])

        conn.autocommit = True

        return conn


    def query(self, queryDict):
        
        self.query_parts_og = queryDict
        result = {}
        
        
        #self._reset()
        '''
        We loops through the query parts to call various functions to perform the translation so it works with this datasource
        '''
        for key in queryDict:
            if key == 'action': 
                continue
            do = f"_{key}"
            
            
            if hasattr(self, do) and callable(func := getattr(self, do)):

                '''
                check to see if the key already exists. For instance, in a select with a where statement, or anything that 
                requires parameters, we populate the 'data' key with those parameters.
                '''
                if key not in self.query_parts:
                    self.query_parts[key] = func(self.query_parts_og[key])

        self.query_parts['action'] = self._action(self.query_parts_og['action'])

        do = f"_action_{self.query_parts_og['action']}"

        command = ''
        if hasattr(self, do) and callable(func := getattr(self, do)):
            command = func()

        if command:
            Console.log(self.query_parts)
            Console.log(f'{command}')
            success, data, meta = self._run(command)
            Console.db(data)
        else:
            Console.log('No database command to run')
            return False, []
            
        return success, data, meta

        

    def _run(self, queryStr):
        cursor = self.connection.cursor(cursor_factory=psycopg2.extras.DictCursor)#NamedTupleCursor)#DictCursor)
        result = {}
        success: bool
        
        try:
            '''
            little check to see if we're working with a single row of data (a dict)
            '''
            if len(self.query_parts['data']) > 0:
                cursor.execute(queryStr, self.query_parts['data'][0])
            else:
                cursor.execute(queryStr)
            
            success = True
        except Exception as err:
            Console.error(f'{err}')
            Console.error(f'{type(err)}')
            success = False

        count = cursor.rowcount
        try:
            columns: list = [col.name for col in cursor.description]
            res = cursor.fetchall()

            count = cursor.rowcount
            data: list = []
            for row in res:
                row_data: dict = {}
                for field in columns:
                    row_data[field] = row[field]
                data.append(row_data)


            result = data
            #result[0][ self.query_parts['pk'][0] ] = id

        except:
            pass

        cursor.close()

        meta:dict = {}
        meta['count'] = count
        return success, result, meta    
        

    def _process_field(self):
        pass


    def _action(self, value):
        match value:
            case 'select':
                return 'SELECT'
            case 'update':
                return 'UPDATE'
            case 'insert':
                return 'INSERT INTO'
            case 'delete':
                return 'DELETE'
            case 'create':
                return 'CREATE TABLE'
            case 'drop':
                return 'DROP TABLE'
            case 'alter':
                return 'ALTER TABLE'


    def _field(self, fieldArr):
        if not fieldArr:
            return ''

        fieldArr = [f'"{field}"' for field in fieldArr]

        string = ','.join(fieldArr)
        return string


    def _order(self, value: list):
        if not value:
            return ''

        orderArr: list = []
        for field, direction in value:
            orderArr.append(f'"{field}" {direction}')
        
        
        return 'ORDER BY ' + ','.join(orderArr)


    def _pagination(self, value):
        if not value:
            return ''

        limit = value["limit"]

        return(f'LIMIT {limit}')


    def _table(self, value):
        return(f'{value}')


    def _where(self, whereList: list):
        if not whereList:
            return ''
        
        whereArr: list = []
        length = len(whereList)
        counter = 0

        if 'data' in self.query_parts:
            self.query_parts['data'].append({})
        else:
            self.query_parts['data'] = [{}]

        for field, value, operation, group_operation in whereList:

            ## exclude the final group_operation so it doesn't interfere with the remainder of our query after the WHERE clause
            if counter < length-1:
                group_operation = self._translate(group_operation)
            else:
                group_operation = ''

            if len(whereList) == 1:
                counter = ''
            
            if value is None:
                whereArr.append(f'("{field}" IS NULL {group_operation})')
                continue

            whereArr.append(f'("{field}" {self._translate(operation)} %({field}{counter})s) {group_operation}')
            data_field_name = f'{field}{counter}'
            
            self.query_parts['data'][0][f'{field}{counter}'] = value

            if type(counter) == int:
                counter = counter + 1

        return 'WHERE ' + ' '.join(whereArr)


    def _database(self, value):
        return(f'{value}')


    def _data(self, value):
        return value


    def _pk(self, value):
        return [f'"{field}"' for field in value]


    def _action_create(self):
        arr = []
        for field, schema in self.query_parts_og['field'].items():

            # build up each column with various attributes like, auto, pk, not null etc
            line = [f'"{field}"']

            if 'auto' in schema.keys():
                line.append('SERIAL PRIMARY KEY')
            else:
                line.append(f'{schema["type"].upper()}')

            if schema['type'] not in 'int,number':
                line.append(f'({str(schema["size"])})')
       
            if 'required' in schema.keys():
                line.append('NOT NULL')
            line.append(',')
            
            arr.append(' '.join(line))
        
        result = ''.join(arr)[:-1]
        self.query_parts['field'] = '(' + result + ')'
        
        return self.query_parts['action'] + ' ' + self.query_parts['table'] + ' ' + self.query_parts['field'] + ' '
        

    def _action_drop(self):
        return self.query_parts['action'] + ' ' + self.query_parts['table']


    def _action_select(self):
        return self.query_parts['action'] + ' ' + self.query_parts['field'] + ' FROM ' + self.query_parts['table'] + ' ' + self.query_parts['where'] + ' ' + self.query_parts['order'] + ' ' + self.query_parts['pagination']


    def _action_insert(self):
        valueArr = []
        Console.error(self.query_parts)
        for field, value in self.query_parts['data'][0].items():
            valueArr.append(f'%({str(field)})s')

        return self.query_parts['action'] + ' ' + self.query_parts['table'] + ' (' + self.query_parts['field'] + ') VALUES (' + ','.join(valueArr) + ') RETURNING ' + ','.join(self.query_parts['pk'])


    def _action_update(self):
        if len(self.query_parts['data']) == 0:
            return False
            
        rowArr = []
        for field, value in self.query_parts['data'][0].items():
            if field not in self.query_parts_og['pk']:
                rowArr.append(f'"{field}" = %({str(field)})s')

        return self.query_parts['action'] + ' ' + self.query_parts['table'] + ' SET ' + ','.join(rowArr) + ' ' + self.query_parts['where']


    def _action_delete(self):
        return self.query_parts['action'] + ' FROM ' + self.query_parts['table'] + ' ' + self.query_parts['where']


    def _reset(self):
        self.query_parts = self.query_parts_og


    '''
    converts keywords to this database specific keywords
    '''
    def _translate(self, value):

        return value


import inspect
import os
class Color:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\33[101m'
    END = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    GREY = '\33[90m'
    RED = '\033[91m'

class Console:

    @classmethod
    def stack(cls):
        frame = inspect.stack()[2]
        filepath= frame.filename
        lineno = frame.lineno
        function = frame.function
        filename = os.path.basename(filepath)

        return(f'{Color.HEADER}{Color.BOLD}{filename}:{function}:{lineno}{Color.END}')


    @classmethod
    def log(cls, string):
        stack = Console.stack()
        print(f"{stack}: {Color.OKCYAN}{string}{Color.END}")
        

    @classmethod
    def warn(cls, string):
        stack = Console.stack()
        print(f"{stack}: {Color.WARNING}{string}{Color.END}")

    @classmethod
    def error(cls, string):
        stack = Console.stack()
        print(f"{Color.FAIL}{stack}: {string}{Color.END}")
    
    @classmethod
    def fail(self, string):
        self.error(string)

    @classmethod
    def ok(cls, string):
        stack = Console.stack()
        print(f"{stack}: {Color.OKGREEN}{string}{Color.END}")

    @classmethod
    def success(self, string):
        stack = Console.stack()
        self.ok(string)

    @classmethod
    def info(cls, string):
        
        stack = Console.stack()
        print(f"{stack}: {Color.OKCYAN}{string}{Color.END}")

    # formatting for data retrieval
    @classmethod
    def db(cls, string):
        stack = Console.stack()
        print(f"{stack}: {Color.GREY}{string}{Color.END}")

    @classmethod
    def header(cls, string):
        print(f"{Color.HEADER}{string}{Color.END}")


from .util import Console
from .validation import Validation

from importlib import import_module

from .dictionary import Dictionary

class Datasource(Dictionary, Validation):

    queryParts:dict = {
        'field':{},
        'pagination':{},
        'where':[],
        'order':[],
        'action':'',
        'table': '',
        'database': '',
        'pk':{},
        'data':[]
    }

    @property
    def error(self):
        return self._error

    
    @error.setter
    def error(self, errorString):
        self._error.append(errorString)


    @property
    def data(self):
        return self._data


    @data.setter
    def data(self, dataInput: dict or list):

        if type(dataInput) is dict:
            data = [dataInput]
        else:
            data = dataInput

        #dataList: list = [*dataDict]
        fieldList = [*data[0],]
        self.queryParts["field"] = Validation.fieldExists(self, fieldList)
        
        self.queryParts['data'] = []

        for row in data:
            dataDict, error = Validation.check(self, row)
        
            if error:
                self.error = error
                continue

            temp = {}
            rowData = {}
            for field in self.queryParts["field"]:
                #self._data[field] = dataDict[field]
                #self.queryParts['data'][field] = dataDict[field]
                rowData[field] = dataDict[field]
                temp[field] = dataDict[field]

            self.queryParts['data'].append(temp)
            
            self._data.append(rowData)


    def __init__(self, schemaFile, connectionInfo: dict):
        Console.info('Datasource instance.................')
        super().__init__(schemaFile)
        
        database_type = self.database_spec['type']

        self.db = import_module(f'.module.connection.{database_type}', package='ivyorm').Connection(connectionInfo)
        
        self._data: list = []
        self._error: list = []
        self.id: any
        self.count: int = 0

        self.reset()
        

    def select(self):
        self.queryParts['action'] = 'select'

        if not self.error:
            Console.log(self.queryParts)
            success, result, meta = self.db.query(self.queryParts)
            
            self.reset()
            self._data = result
            
            return success

        self.reset()
        return False


    def insert(self, dataInput: dict or list = None):
        self.queryParts['action'] = 'insert'
        h = dataInput

        if dataInput:
            if type(dataInput) is list:
                self.data = dataInput
            
            if type(dataInput) is dict:
                self.data = dataInput

        if not self.error:
            #self.queryParts['data'] = self.data

            success, result, meta  = self.db.query(self.queryParts)
            self.id = result[0][self.queryParts['pk'][0]]
            #self.data[0][ self.queryParts['pk'][0] ] = self.id
            self.reset()
            return success

        self.reset()
        return False


    def update(self, dataInput: dict or list = None):
        self.queryParts['action'] = 'update'

        if dataInput:
            self.data = dataInput

        data = self._data[0]
        pkList = self.queryParts['pk']
        
        counter = 0
        no_of_pk = len(pkList)
        
        where = {}
        for pk in pkList:

            for field in data.keys():

                if field == pk:
                    counter = counter + 1
                    where[field] = self.data[0][field]


        '''
        all of the pk requirements have not been met and we have no where
        '''
        if no_of_pk != counter and self.queryParts['where'] is None:
            self.error = 'No PK passed in'
            return None

        if no_of_pk == counter:
            for field, value in where.items():
                items = [field, value, '=']
                self.where(items)

        success, result, meta  = self.db.query(self.queryParts)
        self.reset()

        return success


    def delete(self):
        self.queryParts['action'] = 'delete'
        
        if not self.queryParts['where']:
            self.error = 'arguments are a requirement for a DELETE statement'

        success, result, meta = self.db.query(self.queryParts)
        self.reset()
        
        return success


    def field(self, fields: list = None):
        if not fields:
            self.queryParts["field"] = self.field_spec.keys()
            return self

        self.queryParts["field"] = Validation.fieldExists(self, fields)

        return self
    

    '''
    Expects a list of lists
    '''
    def order(self, fieldArr: list):
        for field, direction in fieldArr:
            
            if not Validation.fieldExists(self, field):
                continue

            if direction not in ['DESC','desc','DSC','dsc','ASC','asc']:
                continue

            stuff_to_add = [field, direction]
            self.queryParts["order"].append(stuff_to_add)

        return self


    def where(self, fieldArr: list):
        '''
        fieldArr is a list, these sub lists will at minimum contain 2 values, up to 4 values
        we access the 3rd and 4th value with *args, but need to test for them and assign defaults if
        they don't exist
        '''
        field, value, *args = fieldArr

        operation = '='
        group_operation = 'AND'
        if args:
            if len(args) == 1:
                tmp = super().translate(args[0])
                if tmp:
                    operation = tmp
                else:
                    Console.warn(f'Operation symbol {args[0]} not found. Defaulting to {operation}')

            if len(args) == 2:
                tmp = super().translate(args[1])
                if tmp:
                    group_operation = tmp
                else:
                    Console.warn(f'Group operation symbol {args[1]} not found. Defaulting to {group_operation}')

        stuff_to_add = [field, value, operation, group_operation]
        self.queryParts["where"].append(stuff_to_add)

        return self


    def limit(self, value):
        self.queryParts["pagination"]["limit"] = value
        return self    


    def database(self, database: str):
        self.queryParts["database"] = database


    def table(self, table: str):
        self.queryParts["table"] = table


    def pk(self, pkList: list):
        self.queryParts['pk'] = pkList


    def object(self, object):
        self.table(object)


    def create(self):
        self.queryParts["action"] = 'create'

        fields = {}
        for field in self.field_spec:
            fields[field] = self.field_spec[field]['back']

        self.queryParts["field"] = fields
        self.db.query(self.queryParts)
        self.reset()


    def drop(self):
        self.queryParts["action"] = 'drop'
        self.db.query(self.queryParts)
        self.reset()

    

    '''
    Resets the queryParts dict to blank all the keys
    '''
    def reset(self):
        for key in self.queryParts:
            if type(self.queryParts[key]) in [list, tuple, dict]:
                self.queryParts[key].clear()

        


        self.field()
        self.database(self.database_spec['name'])
        self.table(self.table_spec['name'])
        self.pk(self.table_spec['pk'])


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ivyorm',
 'ivyorm.module',
 'ivyorm.module.connection',
 'ivyorm.module.interface',
 'ivyorm.module.util']

package_data = \
{'': ['*'], 'ivyorm': ['ivypy.egg-info/*', 'model/*']}

setup_kwargs = {
    'name': 'ivyorm',
    'version': '0.1.23',
    'description': 'An ORM for Python',
    'long_description': "Ivy ORM allows you to write data access instructions that are then converted to the database you are working with.\n\nBy providing a schema file to the Datasource class, you are able to (hopefully) quickly write some code that allows you to query a database\n\n# Current status\nThis is a very 'alpha' build. I've only done the very basics such as create and drop tables, and CRUD operations.\nThere is also only one connector - Postgres!",
    'author': 'James Randell',
    'author_email': 'jamesrandell@me.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)
